@@ -1,4 +1,4 @@
-# Copyright 2018 Venelin Mitov
+# Copyright 2016-2019 Venelin Mitov, Georgios Asimomitis
 #
 # This file is part of PCMBase.
 #
@@ -31,12 +31,12 @@ PCMCond.DOU <- function(
 
   Theta <- if(is.Global(model$Theta)) as.vector(model$Theta) else model$Theta[, r]
 
-  Sigma_x <- if(is.Global(model$Sigma_x)) as.matrix(model$Sigma_x) else as.matrix(model$Sigma_x[,, r])
+  Sigma_x <- GetSigma_x(model, "Sigma", r)
 
   Sigma <- Sigma_x %*% t(Sigma_x)
 
   if(!is.null(model$Sigmae_x)) {
-    Sigmae_x <- if(is.Global(model$Sigmae_x)) as.matrix(model$Sigmae_x) else as.matrix(model$Sigmae_x[,,r])
+    Sigmae_x <- GetSigma_x(model, "Sigmae", r)
     Sigmae <- Sigmae_x %*% t(Sigmae_x)
   } else {
     Sigmae <- NULL
@@ -72,8 +72,8 @@ PCMDescribeParameters.DOU <- function(model, ...) {
     H1 = "adaptation rate matrix",
     H2 = "decorrelation rate matrix",
     Theta = "long-term optimum",
-    Sigma_x = "Choleski factor of the unit-time variance rate",
-    Sigmae_x = "Choleski factor of the non-heritable variance or the variance of the measurement error")
+    Sigma_x = "factor of the unit-time variance rate",
+    Sigmae_x = "factor of the non-heritable variance or the variance of the measurement error")
 }
 
 #' @export
@@ -164,9 +164,9 @@ PCMSpecify.DOU <- function(model, ...) {
     Theta = structure(0.0, class = c('VectorParameter'),
                       description = 'long-term optimum'),
     Sigma_x = structure(0.0, class = c('MatrixParameter', '_UpperTriangularWithDiagonal', '_WithNonNegativeDiagonal'),
-                        description = 'Choleski factor of the unit-time variance rate'),
+                        description = 'factor of the unit-time variance rate'),
     Sigmae_x = structure(0.0, class = c('MatrixParameter', '_UpperTriangularWithDiagonal', '_WithNonNegativeDiagonal'),
-                         description = 'Choleski factor of the non-heritable variance or the variance of the measurement error'))
+                         description = 'factor of the non-heritable variance or the variance of the measurement error'))
   attributes(spec) <- attributes(model)
   if(is.null(names(spec))) names(spec) <- c('X0', 'H1', 'H2', 'Theta', 'Sigma_x', 'Sigmae_x')
   if(any(sapply(spec, is.Transformable))) class(spec) <- c(class(spec), '_Transformable')
